SHAKE  {
	var <portID, <port, <server;
	var <streamRoutine, <dataString, <collect = false;
	var lastaccelX, lastaccelY, lastaccelZ;
	var <accelX, <accelY, <accelZ, <accel, <integratedAccel, <integrateA;
	var <gyroX, <gyroY, <gyroZ, <gyroXSum = 0, <gyroYSum = 0, <gyroZSum = 0;
	var <magX, <magY, <magZ, <mag;
	var <compass, <cs0, <cs1, <twog;
	var <navswitch = 0, <>navUpAction, <>navDownAction, <>navPressAction, <>navRelAction;
	var <csPower, <accelPower, <gyroPower, <vibPower, <magPower, <tempPower, <analogPower, 
		<navPower;
	var <accelXBus, <accelYBus, <accelZBus, <accelBus, <integratedAccelBus;
	var <gyroXBus, <gyroYBus, <gyroZBus, <gyroXSumBus, <gyroYSumBus, <gyroZSumBus;
	var <magXBus, <magYBus, <magZBus, <magBus;
	var <compassBus, <cs0Bus, <cs1Bus, <navBus;
	var cperfun, msgs, msgsStream;
	var <accelTrigs, <connected = false;
	
	
	*new {arg portID, server;
		^super.newCopyArgs(portID).init(server);
		}
		
	init {arg argServer;
		// to be used for control bus writing and control bus allocation
		server = argServer ?? {Server.default};
		twog = true; // assume 2g? need to figure out a way to save parameters
		// init the busses to nil;
		accelXBus = accelYBus = accelZBus = accelBus = gyroXBus = gyroYBus = gyroZBus = 
		gyroXSumBus = gyroYSumBus = gyroZSumBus = magXBus = magYBus = magZBus = compassBus = 
		cs0Bus = cs1Bus = navBus = integratedAccelBus = nil;
		cperfun = {
			Routine.run({
				this.closeStream; this.stopSendStream; this.collectStream; this.sendStream;
				})
			};
		integrateA = Integrate(0.0, 0.95);
		msgs = [];
		}
	
	*findSHAKE {
		var devs, shakes;
		shakes = [];
		devs = SerialPort.devices;
		devs.do({arg me;
			me.contains("SHAKE").if({
				shakes = shakes.add(me);
				});
			});
		^shakes;
		}
	
	*resetToFacSettings {arg dev;
		var cmds;
		cmds = ["$WRI,0008,00*00", // ACCEL
			"$WRI,0009,00*00", // GYRO          
			"$WRI,000a,00*00", // MAH           
			"$WRI,000b,00*00", // CH   
			"$WRI,000c,00*00", // CS0     
			"$WRI,000d,00*00", // CS1     
			"$WRI,000e,00*00", // AIN0   
			"$WRI,000f,00*00", // AIN1
			"$WRI,0000,FB*00", // Pwr Ctrl
			"$WRI,0014,01*00", // A Filt
			"$WRI,0015,01*00", // G Filt
			"$WRI,0016,01*00", // M Filt
			"$WRI,0018,01*00", // CS0 Filt
			"$WRI,0019,01*00", // CS1 Filt
			"$WRI,001A,01*00", // AIN0 Filt
			"$WRI,001B,01*00", // AIN1 Filt
			"$WRI,0002,00*00", // ODF
			"$WRI,0004,00*00", // ACC 
			"$WRI,0008,05*00", // ACC ODR 
			"$WRI,0012,00", // Streaming
			"$WRI,0007,00", // X-Axes En
			"$WRI,0006,00", // Data Cal Mod
			"$WRI,0010,02", // Gyro OS TC enable
			"$WRI,001E,00", // A01 off
			"$WRI,003A,00", // H01 off
			"$VIB,01,00,00,00,FF,03,00,30,FF,03,01,30~",
			"$VIB,02,00,00,00,FF,03,00,30,FF,03,01,30,FF,03,00,30,FF,03,01,30,FF,03,00,30,FF,03,01,30~",
			"$VIB,03,00,00,00,FF,03,00,30,FF,03,01,30,FF,03,00,30,FF,03,01,30,FF,03,00,30,FF,03,01,30~",
			"$VIB,04,00,00,00,FF,03,00,30,FF,03,01,30~"
			];
		"Resetting the SHAKE to factory settings... this may take a moment".postln;
		Routine.run({
			var port;
			port = SerialPort(dev, baudrate: 230400);
			cmds.do({arg me;
				port.putAll(me);
				});
			port.close;
			"SHAKE has been reset to factory settings".postln;
			})
		}
		
	*initForSC {arg dev;
		var cmds;
		cmds = [
			"$WRI,0002,00", // set to ASCII
			"$WRI,0008,00", // set accel output to 0
			"$WRI,0009,00", // set gyro
			"$WRI,000A,00", // set mag
			"$WRI,000B,00", // set cs
			"$WRI,000C,00", // set compass
			"$WRI,000D,00"]; // set to ASCII
		Routine.run({
			var port;
			port = SerialPort(dev, baudrate: 230400);
			cmds.do({arg me;
				port.putAll(me);
				});
			port.close;
			"SHAKE has been calibrated for SC".postln;
			})
		}		
		
	/* BASIC FUNCTIONS */
	connect {
		Routine.run({
			// warning! halts language evaluation while opening... plan ahead for this!
			try({
				port = SerialPort(portID, baudrate: 230400);
				// pull up the splash screen
				209.do({arg i;
					port.read.asAscii.post;
					});
				// no checksum, send ASCII
				1.wait;
				this.initConfig;
				this.collectStream;
				connected = true;
				//^true;
				}, {
				"\nWARNING: SHAKE couldn't connect... check connections".postln;
				connected = false;
				})
			});
		}
	
	sendStream {
		msgsStream = Routine.run({
			loop({
				(msgs.size > 0).if({
					msgs.do({arg me;
						port.putAll(me);
						})
					});
				msgs = [];
				0.001.wait;
				});
			})
		}
		
	stopSendStream {
		msgsStream.stop;
		msgs = [];
		}
		
	setASCIIData {
		this.sendMsg("$WRI,0002,00");
		}
		
	powerAll {
		csPower=accelPower=gyroPower=vibPower=magPower=tempPower=analogPower=navPower=true;		this.sendPowerMsg;
		}
	
	sendPowerMsg {
		var power;
		power = 0;
		power = power.setBit(0, csPower);
		power = power.setBit(1, accelPower);
		power = power.setBit(2, gyroPower);		
		power = power.setBit(3, vibPower);
		power = power.setBit(4, magPower);
		power = power.setBit(5, tempPower);
		power = power.setBit(6, analogPower);
		power = power.setBit(7, navPower);
		this.sendMsg("$WRI,0000,"++power.asHexString(2));
		}

	calibrate {arg sensor;
		this.sendMsg("$WRI,010A,"++sensor.asHexString(2))
		}
				
	requestSplash {
		this.sendMsg("$WRI,0101,01")
		}
		
	initConfig {
		this.setASCIIData;	
		this.setAccelRate(0);
		this.setGyroRate(0);
		this.setMagRate(0);
		this.setCompassRate(0);
		this.setCs0Rate(0);
		this.setCs1Rate(0);
	}
	
	close {
		Routine.run({
			collect.if({
				this.closeStream;
				});
			1.wait;
			port.close;
			connected = false;
			"Closed".postln
			})
		}
	
	// this will collect packets from the SHAKE as other methods request them. 
	// It will parse the ASCII, toss out the ack, then send to the appropriate method
	// for parsing and storing values in instance vars.
	// Bascilly just waits for messages, and parses them
	collectStream {
		var thisByte, tmpString, test;
		collect.not.if({
			this.sendStream;
			streamRoutine = Routine.run({
				collect = true;
				tmpString = "";
				CmdPeriod.add(cperfun);
				loop({
					thisByte = port.read;
					// check if there is a new line. If there is
					// close off the string, parse it, and send it to the 
					// appropriate place
					(thisByte == 0x0A).if({
						dataString = tmpString;
						// debug
						// dataString.postln;
						test = dataString[0..3];
						case
							{test == "$ACK"}
							{dataString.postln}
							// accel packet
							{test == "$ACC"}
							{this.parseAccel(dataString)}
							// gyro (or Angular Rate Sensor)
							{test == "$ARS"}
							{this.parseGyro(dataString)}
							// magnometer
							{test == "$MAG"}
							{this.parseMag(dataString)}
							{test == "$HED"}
							{this.parseCompass(dataString)}
							{test == "$CS0"}
							{this.parseCS(dataString, 0)}
							{test == "$CS1"}
							{this.parseCS(dataString, 1)}
							{test == "$NVU"}
//							{navswitch = 1; this.setnav; navUpAction.value}
							{this.setnav(1)}
							{test == "$NVD"}
//							{navswitch = -1; this.setnav; navDownAction.value}
							{this.setnav(-1)}
							{test == "$NVN"}
//							{navswitch = 0; this.setnav; navRelAction.value}
							{this.setnav(0)}
							{test == "$NVC"}
//							{navswitch = 2; this.setnav; navPressAction.value}
							{this.setnav(2)}
							{true}
							{dataString.postln};
						tmpString = "";
						}, {
						tmpString = tmpString ++ thisByte.asAscii;
						})
					})
				});
			})
		}
		
	sendMsg {arg string;
		msgs = msgs.add(string);
		}
	
	readSign {arg char;
		^(char == $-).if({-1}, {1});
		}
	
	setControl {arg bus, val;
		bus.notNil.if({
			server.sendMsg(\c_set, bus, val);
			})
		}
		
	checkBus {arg bus;
		bus = bus.asUGenInput;
		^bus ?? server.controlBusAllocator.alloc(1);
		}
		
	// seems to be a problem if there is data still in the stream
	// does this relieve the pressure?
	flushStream {arg turn = 0;
		var val;
		//turn.postln;
		val = port.next;
		//val.postln;
		((turn < 10000) && val.notNil).if({
			this.flushStream(turn + 1)
			});
		}
			
	closeStream {
		this.initConfig;
		this.stopSendStream;
		streamRoutine.stop;
		this.flushStream(0);
		collect = false;
		CmdPeriod.remove(cperfun);
		}
			
	/* ACCEL FUNCTIONS */
	
	// set to false for off, true for on
	accelPower_ {arg power = true; 
		accelPower = power;
		this.sendPowerMsg;
		}
	
	accelXToControl {arg aBus;
		accelXBus = this.checkBus(aBus);
		this.setControl(accelXBus, accelX);
		}

	accelYToControl {arg aBus;
		accelYBus = this.checkBus(aBus);
		this.setControl(accelYBus, accelY);
		}
		
	accelZToControl {arg aBus;
		accelZBus = this.checkBus(aBus);
		this.setControl(accelZBus, accelZ);
		}
		
	accelSumToControl {arg aBus;
		accelBus = this.checkBus(aBus);
		this.setControl(accelBus, accel);
		}
		
	accelIntegrateToControl {arg aBus;
		integratedAccelBus = this.checkBus(aBus);
		this.setControl(integratedAccelBus, integratedAccel);
		}
				
	// route the values to control busses .. if nil
	accelToControl {arg xBus, yBus, zBus, accelBus, integratedAccelBus;
		this.accelXToControl(xBus);
		this.accelYToControl(yBus);
		this.accelZToControl(zBus);
		this.accelSumToControl(accelBus);	
		this.accelIntegrateToControl(integratedAccelBus);	
		}
		
	setAccelRate {arg newrate;
		this.sendMsg("$WRI,0008,"++newrate.asHexString(2));
		}
		
	addAccelTrigger {arg id, limit = 0, testfun, triggerfun, active = true;
		accelTrigs.isNil.if({
			accelTrigs = Dictionary.new;
			});
		// limit, testfunm triggerfun, lasttrig
		accelTrigs.put(id, [limit, testfun, triggerfun, 0.0, active]);
		}
	
	activateAccelTrigger {arg id;
		accelTrigs[id][3] = 0.0; // reset so a trigger can happen NOW!
		accelTrigs[id][4] = true;
		}
		
	deactivateAccelTrigger {arg id;
		accelTrigs[id][4] = false;
		}
	
	removeAccelTrigger {arg id;
		accelTrigs.removeAt(id)
		}
			
	// this gs is 0 or 1 (for +-2g and +-6g)
	// hpf is 0 (for off) or 1 (for on)
	setAccelParams {arg gs, hpf = 0;
		(gs == 0).if({twog = true}, {twog = false});
		this.sendMsg("$WRI,0004,"++(gs + (hpf * 2)).asHexString(2));
		}
		
	requestAccel {
		collect.if({
			this.sendMsg("$WRI,0100,01");
			})
		}
	
	parseAccel {arg string;
		var xg, yg, zg, xsign, ysign, zsign, tmpx, tmpy, tmpz, curstate;
		var limit, test, trigfunc, lasttrig, active, now;
		xg = string[6..9].interpret;
		yg = string[12..15].interpret;
		zg = string[18..21].interpret;
		xsign = this.readSign(string[5]);
		ysign = this.readSign(string[11]);
		zsign = this.readSign(string[17]);
		tmpx = (xg * xsign * 0.001);
		tmpy = yg * ysign * 0.001; 
		tmpz = zg * zsign * 0.001;
		lastaccelX.isNil.if({
			lastaccelX = tmpx;
			lastaccelY = tmpy;
			lastaccelZ = tmpz;
			});
		accelX = tmpx-lastaccelX;
		accelY = tmpy-lastaccelY;
		accelZ = tmpz-lastaccelZ;
		lastaccelX = tmpx;
		lastaccelY = tmpy;
		lastaccelZ = tmpz;
		accel = accelX + accelY + accelZ;
		integratedAccel = integrateA.onePole(accel.abs);
		// send to controls if control is notNil
		this.setControl(accelXBus, accelX);
		this.setControl(accelYBus, accelY);
		this.setControl(accelZBus, accelZ);
		this.setControl(accelBus, accel);	
		this.setControl(integratedAccelBus, integratedAccel);
		// check the trigger function
		accelTrigs.do({arg me;
			#limit, test, trigfunc, lasttrig, active = me;
			(active and: {test.value(accel, accelX, accelY, accelZ)}).if({
				now = Main.elapsedTime;
				((now - lasttrig) > limit).if({
					trigfunc.value(accel, accelX, accelY, accelZ);
					me[3] = now;
					})
				})
			})
				
//		accelTriggerFun.notNil.if({
//			(accelTriggerTime.isNil or: 
//				{(Main.elapsedTime - accelTriggerTime) > accelTriggerLimit}).if({
//					curstate = accelTriggerTest.value;
//					curstate.if({
//						accelTriggerTime = Main.elapsedTime;
//						accelTriggerFun.value;
//						})
//					})
//				});
		}
	
//	zeroSpeed {
//		speed = 0;
//		}
		
	calibrateAccel {
		this.calibrate(2);
		}
	
	/* GYRO FUNCTIONS */

	gyroPower_ {arg power = true; 
		gyroPower = power;
		this.sendPowerMsg;
		}
	
	gyroXToControl {arg aBus;
		gyroXBus = this.checkBus(aBus);
		this.setControl(gyroXBus, gyroX);
		}
		
	gyroYToControl {arg aBus;
		gyroYBus = this.checkBus(aBus);
		this.setControl(gyroYBus, gyroY);
		}
		
	gyroZToControl {arg aBus;
		gyroZBus = this.checkBus(aBus);
		this.setControl(gyroZBus, gyroZ);
		}
	
	gyroXSumToControl {arg aBus;
		gyroXSumBus = this.checkBus(aBus);
		this.setControl(gyroXSumBus, gyroXSum);
		}

	gyroYSumToControl {arg aBus;
		gyroYSumBus = this.checkBus(aBus);
		this.setControl(gyroYSumBus, gyroYSum);
		}

	gyroZSumToControl {arg aBus;
		gyroZSumBus = this.checkBus(aBus);
		this.setControl(gyroZSumBus, gyroZSum);
		}
	
	gyroToControl {arg gyroXBus, gyroYBus, gyroZBus, gyroXSumBus, gyroYSumBus, gyroZSumBus; 
		this.gyroXToControl(gyroXBus);
		this.gyroYToControl(gyroYBus);
		this.gyroZToControl(gyroZBus);	
		this.gyroXSumToControl(gyroXSumBus);
		this.gyroYSumToControl(gyroYSumBus);
		this.gyroZSumToControl(gyroZSumBus);
		}
							
	setGyroRate {arg newrate;
		this.sendMsg("$WRI,0009,"++newrate.asHexString(2));
		}
	
	setGyroTempCompensation {arg bool = false;
		var val;
		val = bool.if({2}, {0});
		this.sendMsg("WRI,0010,"++val.asHexString(2));
		}
				
	requestGyro {
		this.sendMsg("$WRI,0100,02")
		}
		
	parseGyro {arg string;
		var xg, yg, zg, xsign, ysign, zsign, rscale;
		rscale =  0.00010893246187364; // 1 / 9180
		xg = string[6..9].interpret;
		yg = string[12..15].interpret;
		zg = string[18..21].interpret;
		xsign = this.readSign(string[5]);
		ysign = this.readSign(string[11]);
		zsign = this.readSign(string[17]);		
		gyroX = xg * xsign * rscale;
		gyroY = yg * ysign * rscale;
		gyroZ = zg * zsign * rscale;
		gyroXSum = gyroXSum + gyroX;
		gyroYSum = gyroYSum + gyroY;
		gyroZSum = gyroZSum + gyroZ;
		this.setControl(gyroXBus, gyroX);
		this.setControl(gyroYBus, gyroY);
		this.setControl(gyroZBus, gyroZ);		
		this.setControl(gyroXSumBus, gyroXSum);
		this.setControl(gyroYSumBus, gyroYSum);
		this.setControl(gyroZSumBus, gyroZSum);
		}
	
	zeroGyroXSum {
		gyroXSum = 0.0;
		}
		
	zeroGyroYSum {
		gyroYSum = 0.0;
		}
		
	zeroGyroZSum {
		gyroZSum = 0.0;
		}
			
	caibrateGyro {
		this.calibrate(3);
		}
		
	/* Magnometer functions */
		
	magPower_ {arg power = true;
		magPower = power;
		this.sendPowerMsg;
		}

	magXToControl {arg aBus;
		magXBus = this.checkBus(aBus);
		this.setControl(magXBus, magX);
		}

	magYToControl {arg aBus;
		magYBus = this.checkBus(aBus);
		this.setControl(magYBus, magY);
		}
		
	magZToControl {arg aBus;
		magZBus = this.checkBus(aBus);
		this.setControl(magZBus, magZ);
		}
		
	magSumToControl {arg aBus;
		magBus = this.checkBus(aBus);
		this.setControl(magBus, mag);
		}
				
	// route the values to control busses .. if nil
	magToControl {arg xBus, yBus, zBus, magBus;
		this.magXToControl(xBus);
		this.magYToControl(yBus);
		this.magZToControl(zBus);
		this.magSumToControl(magBus);	
		}
			
	setMagRate {arg newrate;
		this.sendMsg("$WRI,000A,"++newrate.asHexString(2));
		}
		
	requestMag {
		this.sendMsg("$WRI,0100,04")
		}
		
	parseMag {arg string;
		var xg, yg, zg, xsign, ysign, zsign;
		xg = string[6..9].interpret;
		yg = string[12..15].interpret;
		zg = string[18..21].interpret;
		xsign = this.readSign(string[5]);
		ysign = this.readSign(string[11]);
		zsign = this.readSign(string[17]);
		magX = xg * xsign * 0.001;
		magY = yg * ysign * 0.001;
		magZ = zg * zsign * 0.001;
		mag = xg + yg + zg * 0.001;
		this.setControl(magXBus, magX);
		this.setControl(magYBus, magY);
		this.setControl(magZBus, magZ);		
		this.setControl(magBus, mag);
		}
	
	calibrateMag {
		this.calibrate(1);
		}
		
	/* Compass functions */
	compassToControl {arg compassBus;
		compassBus = this.checkBus(compassBus);
		this.setControl(compassBus, compass);
		}
		
	setCompassRate {arg newrate;
		this.sendMsg("$WRI,000B,"++newrate.asHexString(2));
		}
		
	requestCompass {
		this.sendMsg("$WRI,0100,08")
		}
		
	parseCompass {arg string;
		compass = string[5..8].interpret * 0.1;
		this.setControl(compassBus, compass);
		}	
	
	/* capacitive sensor funtions */	
	csPower_ {arg power = true;
		csPower = power;
		this.sendPowerMsg;
		}
	
	csToControl {arg cs0Bus, cs1Bus;
		this.cs0ToControl(cs0Bus);
		this.cs1ToControl(cs1Bus);
		}
		
	cs0ToControl {arg aBus;
		cs0Bus = this.checkBus(aBus);
		this.setControl(cs0Bus, cs0);
		}

	cs1ToControl {arg aBus;
		cs1Bus = this.checkBus(aBus);
		this.setControl(cs1Bus, cs1);
		}

/* not working */	
//	setCsRate {arg rate;
//		Routine.run({
//			this.setCs0Rate(rate);
//			this.setCs1Rate(rate);
//			})
//		}
		
	setCs0Rate {arg newrate;
		this.sendMsg("$WRI,000C,"++newrate.asHexString(2));
		}
				
	setCs1Rate {arg newrate;
		this.sendMsg("$WRI,000D,"++newrate.asHexString(2));
		}
				
	requestCs0 {
		this.sendMsg("$WRI,0100,10")
		}
		
	requestCs1 {
		this.sendMsg("$WRI,0100,20")
		}
		
	parseCS {arg string, sensorID;
		var tmp;
		tmp = string[5..8].interpret * 0.1;
		(sensorID == 0).if({
			cs0 = tmp;
			this.setControl(cs0Bus, cs0);
			}, {
			cs1 = tmp;
			this.setControl(cs1Bus, cs1);
			});
		}
	
	caibrateCs {
		this.calibrate(6);
		}
		
	/* nav switch controls */
	
	navToControl {arg aBus;
		navBus = this.checkBus(aBus);
		this.setControl(navBus, navswitch);
		}
		
	setnav {arg navswitchVal;
		navswitch = navswitchVal;
		this.setControl(navBus, navswitch);
		case
			{navswitch == 1}
			{navUpAction.do({arg me;
				me.value;
				})
			}
			{navswitch == 0}
			{navRelAction.do({arg me;
				me.value;
				})
			}
			{navswitch == -1}
			{navDownAction.do({arg me;
				me.value;
				})
			}
			{navswitch == 2}
			{navPressAction.do({arg me;
				me.value;
				})
			}
		}
	
	addNavUpAction {arg id, func;
		navUpAction.isNil.if({
			navUpAction = Dictionary.new;
			});
		navUpAction.put(id, func);
		}
		
	removeNavUpAction {arg id;
		navUpAction.notNil.if({
			navUpAction.removeAt(id);
			})
		}

	addNavRelAction {arg id, func;
		navRelAction.isNil.if({
			navRelAction = Dictionary.new;
			});
		navRelAction.put(id, func);
		}
		
	removeNavRelAction {arg id;
		navRelAction.notNil.if({
			navRelAction.removeAt(id);
			})
		}

	addNavDownAction {arg id, func;
		navDownAction.isNil.if({
			navDownAction = Dictionary.new;
			});
		navDownAction.put(id, func);
		}
		
	removeNavDownAction {arg id;
		navDownAction.notNil.if({
			navDownAction.removeAt(id);
			})
		}

	addNavPressAction {arg id, func;
		navPressAction.isNil.if({
			navPressAction = Dictionary.new;
			});
		navPressAction.put(id, func);
		}
		
	removeNavPressAction {arg id;
		navPressAction.notNil.if({
			navPressAction.removeAt(id);
			})
		}
									
	/* vibrotactile controls */
	
	vibPower_ {arg power = true;
		vibPower = power;
		this.sendPowerMsg;
		}
	
	// speed should be between 0 and 1, durs between 0 and 2.55 seconds
	// speed and dur should be comma separated arrays of speed and time
	// e.g 1, [1, 0.2]
	setVibProfile {arg memSlot ... speedAndDur;
		var string, speed, dur;
		((memSlot >= 1) && (memSlot <= 255)).if({
			string = "$VIB,"++memSlot.asHexString(2)++",00,00,00";
			speedAndDur.do({arg me;
				#speed, dur = me;
				speed = "," ++ (speed * 255).asInteger.asHexString(2) ++ ",";
				dur = (dur * 100).asInteger.asHexString(2);
				string = string ++ speed ++ dur;
				});
			this.sendMsg(string++"~");
			}, {
			"memSlot must be between 1 and 255".warn;
			})
		}
	
	// motor must be 0 (main), 1 (left) or 2 (right)
	vibrate {arg motor = 0, memSlot;
		motor = (motor + 2).asHexString(2);
		this.sendMsg("$WRI,01"++motor++","++memSlot.asHexString(2));
		}
	
}